@@ -20,7 +20,9 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
 
-#define RFU_VERSION 10
+#define RFU_LOCAL 1  
+
+
 #include "Basic_utils.h"
 #include <R_ext/Lapack.h>
 #include <R_ext/Linpack.h>
@@ -30,11 +32,20 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "xport_import.h"
 
 
+const char // constants cannot be exported; 
+*KKR_TYPE_NAMES[LAST_R_TYPE_NAME + 1] = { // never change ! see AutoRFU.cc
+  "NILSXP" /* 0 */,
+  "SYMSXP", "LISTSXP", "CLOSXP", "ENVSXP", "PROMSXP",
+  "LANGSXP", "SPECIALSXP", "BUILTINSXP", "CHARSXP", "LGLSXP" /* 10 */,
+  "??", "??", "INTSXP", "REALSXP", "CPLXSXP",
+  "STRSXP", "DOTSXP", "ANYSXP", "ECSXP", "EXPRSXP" /*20 */,
+  "BCODESXP", "EXTPTRSXP", "WEAKREFSXP", "RAWSXP", "S4SXP" /* 25 */,
+  "", "", "", "", "NEWSXP" /* 30 */,
+  "FREESXP", "??SXP"};
+
+
 
-#define USE_OWN_ALGXX(SCALAR_LEN, PARALLEL)				\
-  (PARALLEL < 8 || OPTIONS.basic.LaMaxTakeIntern >= SCALAR_LEN) // OK
 #define USE_OWN_ALG(SCALAR_LEN, PARALLEL) true
-
 #define USE_OWN_SCALAR_PROD true
 
 #define SCALAR(A,B,C) scalarX(A,B,C, SCALAR_AVX)
@@ -47,11 +58,11 @@ void strcopyN(char *dest, const char *src, int n) {
   dest[n] = '\0';
 }
 
-void AtA(double *a, Long nrow, Long ncol, double *C) {
+void AtA(double *a, Long nrow, Long ncol, double *C, int VARIABLE_IS_NOT_USED cores) {
   // C =  A^T %*% A
   if (USE_OWN_ALG(nrow, ncol) || nrow * ncol > MAXINT) {   
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (MULTIMINSIZE(ncol)) schedule(dynamic, 20)
+#pragma omp parallel for num_threads(cores) schedule(dynamic, 20) if (MULTIMINSIZE(ncol)) 
 #endif  
     for (Long i=0; i<ncol; i++) {
       double 
@@ -90,13 +101,13 @@ void xA_noomp(double *x, double*A, Long nrow, Long ncol, double *y) {
 }
 
 
-void xA(double *x, double*A, Long nrow, Long ncol, double *y) {
+void xA(double *x, double*A, Long nrow, Long ncol, double *y, int VARIABLE_IS_NOT_USED cores) {
   if (A == NULL) {
     if (nrow != ncol || nrow <= 0) BUG;
     MEMCOPY(y, x, sizeof(double) * nrow);
   } else {
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (MULTIMINSIZE(ncol) && MULTIMINSIZE(nrow))
+#pragma omp parallel for num_threads(cores) schedule(static) if (MULTIMINSIZE(ncol) && MULTIMINSIZE(nrow))
 #endif  
    for (Long i=0; i<ncol; i++) y[i] = SCALAR(x, A + i * nrow, nrow);
   } 
@@ -119,18 +130,19 @@ void xA(double *x1, double *x2,  double*A, Long nrow, Long ncol, double *y1,
 }  
   
 
-double xAx(double *x, double*A, Long nrow) {
+double xAx(double *x, double*A, Long nrow, int VARIABLE_IS_NOT_USED cores) {
   if (USE_OWN_SCALAR_PROD || nrow * nrow > MAXINT) {
     double sum = 0.0;
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) reduction(+:sum) if (MULTIMINSIZE(nrow) && MULTIMINSIZE(nrow))
+#pragma omp parallel for num_threads(cores) reduction(+:sum) schedule(static) if (MULTIMINSIZE(nrow) && MULTIMINSIZE(nrow))
 #endif  
-    for (Long i=0; i<nrow; i++) sum += x[i] * SCALAR(x, A + i * nrow, nrow);
+    for (Long i=0; i<nrow; i++)
+      sum += x[i] * SCALAR(x, A + i * nrow, nrow);
     return sum;
   } else {
    double alpha = 1.0,
-      beta = 0.0;
-    int incx = 1L;
+     beta = 0.0;
+    int incx = 1;
     double *y = (double*)  MALLOC(nrow * sizeof(double));
     // z = x^\top A
     int nrow0 = nrow;
@@ -142,20 +154,20 @@ double xAx(double *x, double*A, Long nrow) {
   }
 }
 
-void Ax(double *A, double*x, Long nrow, Long ncol, double *y) {
+void Ax(double *A, double*x, Long nrow, Long ncol, double *y, int VARIABLE_IS_NOT_USED cores) {
   if (A == NULL) {
     if (nrow != ncol || nrow <= 0) BUG;
     MEMCOPY(y, x, sizeof(double) * nrow);
   } else {
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (MULTIMINSIZE(ncol) && MULTIMINSIZE(nrow))
+#pragma omp parallel for num_threads(cores) schedule(static) if (MULTIMINSIZE(ncol) && MULTIMINSIZE(nrow))
     for (Long j=0; j<nrow; j++) {
-      double dummy = 0.0;
+      double tmp = 0.0;
       Long k = j;
       for (Long i=0; i<ncol; i++, k+=nrow) { 
-	dummy += A[k] * x[i];
+	tmp += A[k] * x[i];
       }
-      y[j] = dummy;
+      y[j] = tmp;
     }
 #else
     for (Long i=0; i<nrow; i++) y[i]=0.0;
@@ -187,7 +199,8 @@ void Ax(double *A, double*x1, double*x2, Long nrow, Long ncol, double *y1,
 }
 
 
-double XkCXtl(double *X, double *C, Long nrow, Long dim, Long k, Long l) {
+double XkCXtl(double *X, double *C, Long nrow, Long dim, Long k, Long l,
+	      int VARIABLE_IS_NOT_USED cores) {
   // (k-th row of X) * C * (l-th row of X)
   // X is nrow x dim matrix
   // C is dim x dim matrix
@@ -198,7 +211,7 @@ double XkCXtl(double *X, double *C, Long nrow, Long dim, Long k, Long l) {
   Long size = nrow * dim;
   
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) reduction(+:result)
+#pragma omp parallel for num_threads(cores) schedule(static) reduction(+:result)
 #endif
   for (Long j=0; j<size; j+=nrow) {
     double scalar = 0.0;
@@ -210,59 +223,59 @@ double XkCXtl(double *X, double *C, Long nrow, Long dim, Long k, Long l) {
 }
 
 
-void XCXt(double *X, double *C, double *V, Long nrow, Long dim /* dim of C */) {
+void XCXt(double *X, double *C, double *V, Long nrow, Long dim /* dim of C */, int VARIABLE_IS_NOT_USED cores) {
   Long size = nrow * dim;
   double  
     *endpX = X + nrow,
-    *dummy = (double*) MALLOC(sizeof(double) * size); // dummy = XC
-  if (dummy == NULL) RFERROR("XCXt: memory allocation error in XCXt");
+    *tmp = (double*) MALLOC(sizeof(double) * size); // tmp = XC
+  if (tmp == NULL) RFERROR("XCXt: memory allocation error in XCXt");
  
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores)  schedule(static)
 #endif
   for (double *pX = X; pX < endpX; pX++) {
-    double *pdummy = dummy + (pX - X);
+    double *ptmp = tmp + (pX - X);
     for (Long ci=0, cd=0; cd<size; cd+=nrow) {
       double scalar = 0.0;
       for (Long i=0; i<size; i+=nrow) {
         scalar += pX[i] * C[ci++];
       }
-      pdummy[cd] = scalar;
+      ptmp[cd] = scalar;
     }
   }
 
-  // V = dummy X^t
+  // V = tmp X^t
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores)  schedule(dynamic, 20)
 #endif
   for (Long rv=0; rv<nrow; rv++) {
     for (Long cv=rv; cv<nrow; cv++) {
       double scalar=0.0;
       for (Long i=0; i<size; i+=nrow) {
-	scalar += dummy[rv + i] * X[cv + i];
+	scalar += tmp[rv + i] * X[cv + i];
      }
       V[rv + cv * nrow] = V[cv + rv * nrow] = scalar;
     }
   }
 
-  UNCONDFREE(dummy);
+  UNCONDFREE(tmp);
 }
 
 
-double xUy(double *x, double *U, double *y, Long dim) {
+double xUy(double *x, double *U, double *y, Long dim, int VARIABLE_IS_NOT_USED cores) {
   // U a symmetric matrix given by its upper triangular part
   double xVy = 0.0;
   Long    dimM1 = dim - 1;
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (MULTIMINSIZE(dim)) reduction(+:xVy) 
+#pragma omp parallel for num_threads(cores) schedule(static) if (MULTIMINSIZE(dim)) reduction(+:xVy) 
 #endif  
   for (Long d=0; d<dim; d++) {
     Long i, 
       j = dim * d;
-    double dummy = 0.0;
-    for (i=0; i<=d; i++) dummy += x[i] * U[j++];
-    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
-    xVy += dummy * y[d];
+    double tmp = 0.0;
+    for (i=0; i<=d; i++) tmp += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) tmp += x[i] * U[j];
+    xVy += tmp * y[d];
   }
   return xVy;
 }
@@ -273,15 +286,15 @@ double xUy(double *x, double *U, double *y, Long dim) {
   assert(z != NULL);
   Long   dimM1 = dim - 1;
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (MULTIMINSIZE(dim))
+#pragma omp parallel for num_threads(cores) schedule(static) if (MULTIMINSIZE(dim))
 #endif  
   for (Long d=0; d<dim; d++) {
-    double dummy;
+    double tmp;
     Long i,
       j = dim * d;
-    for (dummy = 0.0, i=0; i<=d; i++) dummy += x[i] * U[j++];
-    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
-    if (z!=NULL) z[d] = dummy;
+    for (tmp = 0.0, i=0; i<=d; i++) tmp += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) tmp += x[i] * U[j];
+    if (z!=NULL) z[d] = tmp;
   }
   double xVx;
   SCALAR_PROD(z, x, dim, xVx);
@@ -289,93 +302,97 @@ double xUy(double *x, double *U, double *y, Long dim) {
 
  */
 
-double xUxz(double *x, double *U, Long dim, double *z) {
+double xUxz(double *x, double *U, Long dim, double *z, int VARIABLE_IS_NOT_USED cores) {
  double xVx = 0.0;
   Long dimM1 = dim - 1;
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) reduction(+:xVx)
+#pragma omp parallel for num_threads(cores) schedule(static) reduction(+:xVx)
 #endif
   for (Long d=0; d<dim; d++) {
     Long i, 
       j = dim * d;
-    double dummy = 0.0;
-    for (dummy = 0.0, i=0; i<=d; i++) dummy += x[i] * U[j++];
-    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
-    if (z != NULL) z[d] = dummy;
-    xVx += dummy * x[d];
+    double tmp = 0.0;
+    for (tmp = 0.0, i=0; i<=d; i++) tmp += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) tmp += x[i] * U[j];
+    if (z != NULL) z[d] = tmp;
+    xVx += tmp * x[d];
   }
   return xVx;
 }
 
-double xUx(double *x, double *U, Long dim) {
-    return xUxz(x, U, dim, NULL);
+double xUx(double *x, double *U, Long dim, int VARIABLE_IS_NOT_USED cores) {
+  return xUxz(x, U, dim, NULL, cores);
 }
 
-double x_UxPz(double *x, double *U, double *z, Long dim) {
+double x_UxPz(double *x, double *U, double *z, Long dim, int VARIABLE_IS_NOT_USED cores) {
 // x^t (Ux + z); U dreieckmatrix
   double xVx = 0.0;
   Long    dimM1 = dim - 1;
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) reduction(+:xVx)
+#pragma omp parallel for num_threads(cores) schedule(static) reduction(+:xVx)
 #endif
   for (Long d=0; d<dim; d++) {
     Long i,
       j = dim * d;
-    double dummy = z[d];
-    for (i=0; i<=d; i++) dummy += x[i] * U[j++];
-    for (j += dimM1; i<dim; i++, j+=dim) dummy += x[i] * U[j];
-    xVx += dummy * x[d];
+    double tmp = z[d];
+    for (i=0; i<=d; i++) tmp += x[i] * U[j++];
+    for (j += dimM1; i<dim; i++, j+=dim) tmp += x[i] * U[j];
+    xVx += tmp * x[d];
   }
   return xVx;
 }
 
 
 
-void matmult(double *a, double *b, double *c, Long l, Long m, Long n) {
+void matmult(double *a, double *b, double *c, Long l, Long m, Long n,
+	     int VARIABLE_IS_NOT_USED cores) {
 // multiplying an lxm- and an mxn-matrix, saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores) schedule(static)
 #endif
    for (Long i=0; i<l; i++) {
      double *A = a + i,
        *C = c + i;
      for (Long j=0; j<n; j++) {
-       double dummy = 0.0,
+       double tmp = 0.0,
 	 *B = b + j * m;
-       for (Long k=0; k<m; k++) dummy += A[k*l] * B[k];
-       C[j * l] = dummy;
+       for (Long k=0; k<m; k++) tmp += A[k*l] * B[k];
+       C[j * l] = tmp;
      }
    }
 }
 
 
-double *matrixmult(double *m1, double *m2, Long dim1, Long dim2, Long dim3) {
+double *matrixmult(double *m1, double *m2, Long dim1, Long dim2, Long dim3,
+		   int VARIABLE_IS_NOT_USED cores) {
   double *m0 = (double*) MALLOC(sizeof(double) * dim1 * dim3);
-  matmult(m1, m2, m0, dim1, dim2, dim3);
+  matmult(m1, m2, m0, dim1, dim2, dim3, cores);
   return m0;
 }
 
 
-void Xmatmult(double *A, double *B, double *C, Long l, Long m, Long n) {
+void Xmatmult(double *A, double *B, double *C, Long l, Long m, Long n,
+	      int VARIABLE_IS_NOT_USED cores) {
 // multiplying an lxm- and an mxn-matrix, saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores) schedule(static)
 #endif
   for (Long i=0; i<l; i++) {
     for (Long jl=i, jm=0, j=0; j<n; j++, jl+=l, jm+=m) {
-      double dummy = 0.0;
+      double tmp = 0.0;
       Long endfor = jm + m;
-      for (Long kl=i, k=jm; k<endfor; k++, kl+=l) dummy += A[kl] * B[k]; 
-      C[jl] = dummy;
+      for (Long kl=i, k=jm; k<endfor; k++, kl+=l) tmp += A[kl] * B[k]; 
+      C[jl] = tmp;
     }
   }
 }
 
-void matmulttransposed(double *A, double *B, double *c, Long m, Long l, Long n) {
+void matmulttransposed(double *A, double *B, double *c, Long m, Long l, Long n,
+		       int VARIABLE_IS_NOT_USED cores) {
 // multiplying t(A) and B with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores) schedule(static)
 #endif
   for (Long i=0; i<l; i++) {    
     double *C = c + i,
@@ -390,7 +407,7 @@ void matmulttransposedInt(int *A, int *B, int *c, Long m, Long l, Long n) {
 // multiplying t(A) and B with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores) schedule(static)
 #endif
   for (Long i=0; i<l; i++) {    
     int *C = c + i,
@@ -403,79 +420,83 @@ void matmulttransposedInt(int *A, int *B, int *c, Long m, Long l, Long n) {
 
 
 
-void matmult_2ndtransp(double *a, double *B, double *c, Long l, Long m, Long n) {
+void matmult_2ndtransp(double *a, double *B, double *c, Long l, Long m, Long n,
+		       int VARIABLE_IS_NOT_USED cores) {
 // multiplying A and t(B) with dim(A)=(l, m) and dim(B)=(n, m),
 // saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (l * m * n > 1000)
+#pragma omp parallel for num_threads(cores) schedule(static) if (l * m * n > 1000)
 #endif
   for (Long i=0; i<l; i++) {
     double *C = c + i,
       *A = a + i;
     for (Long j=0; j<n; j++) {
-       double dummy = 0.0,
+       double tmp = 0.0,
 	 *Bj = B + j;
-       for (Long k=0; k<m; k++) dummy += A[k * l] * Bj[k * n];
-       C[j*l] = dummy;
+       for (Long k=0; k<m; k++) tmp += A[k * l] * Bj[k * n];
+       C[j*l] = tmp;
     }
   }
 }
 
 
-void matmult_2ndtransp(double *a, double *B, double *c, Long l, Long m) {
+void matmult_2ndtransp(double *a, double *B, double *c, Long l, Long m,
+		       int VARIABLE_IS_NOT_USED cores) {
 // multiplying A and t(B) with dim(A)=(l, m) and dim(B)=(l, m),
 // saving result in C
   Long lm = l  * m;
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) if (l * m * l > 1000)
+#pragma omp parallel for num_threads(cores) schedule(static) if (l * m * l > 1000)
 #endif
   for (Long i=0; i<l; i++) {
     double *C = c + i,
       *A = a + i;
     for (Long j=0; j<l; j++) {
-       double dummy = 0.0,
+       double tmp = 0.0,
 	 *Bj = B + j;
-       for (Long k=0; k<lm; k+=l) dummy += A[k] * Bj[k];
-       C[j*l] = dummy;
+       for (Long k=0; k<lm; k+=l) tmp += A[k] * Bj[k];
+       C[j*l] = tmp;
     }
   }
 }
 
 
 
-void Xmatmulttransposed(double *A, double *B, double *C, Long m, Long l, Long n) {
+void Xmatmulttransposed(double *A, double *B, double *C, Long m, Long l, Long n,
+			int VARIABLE_IS_NOT_USED cores) {
 // multiplying t(A) and B with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores) schedule(static)
 #endif
   for (Long i=0; i<l; i++) {
     Long im = i * m;
     for (Long jl=i, jm=0, j=0; j<n; j++, jl+=l, jm+=m) {
-      double dummy = 0.0;
+      double tmp = 0.0;
       Long endfor = im + m;
-      for (Long jmk=jm, k=im; k<endfor; k++) dummy += A[k] * B[jmk++]; 
-      C[jl] = dummy;
+      for (Long jmk=jm, k=im; k<endfor; k++) tmp += A[k] * B[jmk++]; 
+      C[jl] = tmp;
     }
   }
 }
 
 
 
-void matmult_tt(double *a, double *B, double *c, Long m, Long l, Long n) {
+void matmult_tt(double *a, double *B, double *c, Long m, Long l, Long n,
+		int VARIABLE_IS_NOT_USED cores) {
 // calculating t(A B) with dim(A)=(m,l) and dim(B)=(m,n),
 // saving result in C
 #ifdef DO_PARALLEL
-#pragma omp parallel for num_threads(RFU_GLOBAL_OPTIONS->basic.cores) 
+#pragma omp parallel for num_threads(cores) schedule(static)
 #endif
   for (Long i=0; i<l; i++) {
     double *A = a + i,
       *C = c + i * l;
     for (Long j=0; j<n; j++) {
-      double dummy = 0.0,
+      double tmp = 0.0,
 	*Bjm = B + j * m;
-      for (Long k=0; k<m; k++) dummy += A[k * l] * Bjm[k];
-      C[j] = dummy;
+      for (Long k=0; k<m; k++) tmp += A[k * l] * Bjm[k];
+      C[j] = tmp;
     }
   }
 }
@@ -515,69 +536,53 @@ SEXP TooSmall(){
 
 
 SEXP Int(int *V, Long n, Long max) {
-  SEXP dummy;
+  SEXP Ans;
   if (V==NULL) return allocVector(INTSXP, 0);
   if (n>max) return TooLarge(n);
-   if (n<0) return TooSmall();
-   PROTECT(dummy=allocVector(INTSXP, n));
-  for (Long i=0; i<n; i++) INTEGER(dummy)[i] = V[i];
+  if (n<0) return TooSmall();   
+  PROTECT(Ans=allocVector(INTSXP, n));
+  MEMCOPY(INTEGER(Ans), V, n * sizeof(int));
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP Int(int* V, Long n) { return Int(V, n, n); }
 
 
 SEXP Logic(bool* V, Long n, Long max) {
-  SEXP dummy;
+  SEXP Ans;
   if (V==NULL) return allocVector(VECSXP, 0);
   if (n>max) return TooLarge(n);
   if (n<0) return TooSmall();
-  PROTECT(dummy=allocVector(LGLSXP, n));
-  for (Long i=0; i<n; i++) LOGICAL(dummy)[i] = V[i];
+  PROTECT(Ans=allocVector(LGLSXP, n));
+  int *ans = LOGICAL(Ans);
+  for (Long i=0; i<n; i++) ans[i] = V[i];
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 SEXP Logic(bool* V, Long n) { return Logic(V, n, n); }
 
 SEXP Num(double* V, Long n, Long max) {
-  SEXP dummy;
+  SEXP Ans;
   if (V==NULL) return allocVector(REALSXP, 0);
   if (n>max) return TooLarge(n);
-   if (n<0) return TooSmall();
-  PROTECT(dummy=allocVector(REALSXP, n));
-  for (Long i=0; i<n; i++) REAL(dummy)[i] = V[i];
+  if (n<0) return TooSmall();
+  PROTECT(Ans=allocVector(REALSXP, n));
+  MEMCOPY(REAL(Ans), V, n * sizeof(double));
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 SEXP Num(double* V, Long n) {  return Num(V, n, n); }
 
-SEXP Result(double* V, Long n, Long max) {
-  SEXP dummy;
-  if (V==NULL) return allocVector(REALSXP, 0);
-  if (n>max) return TooLarge(n);
-   if (n<0) return TooSmall();
- PROTECT(dummy=allocVector(REALSXP, n));
-  for (Long i=0; i<n; i++) REAL(dummy)[i] = (double) V[i];
-  UNPROTECT(1);
-  return dummy;
-}
-
-SEXP Result(double* V, Long n) {
-  return Result(V, n, n);
-}
-
 SEXP Char(const char **V, Long n, Long max) {
-  SEXP dummy;
+  SEXP Ans;
   if (V==NULL) return allocVector(STRSXP, 0);
   if (n>max) return TooLarge(n);
-   if (n<0) return TooSmall();
-   PROTECT(dummy=allocVector(STRSXP, n));
-   for (Long i=0; i<n; i++){
-     SET_STRING_ELT(dummy, i, mkChar(V[i]));  
-   }
+  if (n<0) return TooSmall();
+  PROTECT(Ans=allocVector(STRSXP, n));
+  for (Long i=0; i<n; i++) SET_STRING_ELT(Ans, i, mkChar(V[i]));  
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP Char(const char **V, Long n) { return Char(V, n, n); }
@@ -586,11 +591,11 @@ SEXP Mat(double* V, Long row, Long col, Long max) {
   if (V==NULL) return allocMatrix(REALSXP, 0, 0);
   Long n = row * col;
   if (n>max) return TooLarge(row, col);
-  SEXP dummy;
-  PROTECT(dummy=allocMatrix(REALSXP, row, col));
-  for (Long i=0; i<n; i++) REAL(dummy)[i] = V[i];
+  SEXP Ans;
+  PROTECT(Ans=allocMatrix(REALSXP, row, col));
+  MEMCOPY(REAL(Ans), V, n * sizeof(double));
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP Mat(double* V, Long row, Long col) {
@@ -602,15 +607,14 @@ SEXP Mat_t(double* V, Long row, Long col, Long max) {
   if (V==NULL) return allocMatrix(REALSXP, 0, 0);
   Long n = row * col;
   if (n>max) return TooLarge(row, col);
-  SEXP dummy;
-  PROTECT(dummy=allocMatrix(REALSXP, row, col));
-  for (Long k=0, j=0; j<col; j++) {
-     for (Long i=0; i<row; i++) {
-      REAL(dummy)[k++] = V[j + col * i];
-    }
-  }
+  SEXP Ans;
+  PROTECT(Ans=allocMatrix(REALSXP, row, col));
+  double *ans = REAL(Ans);
+  for (Long k=0, j=0; j<col; j++)
+     for (Long i=0; i<row; i++)
+       ans[k++] = V[j + col * i];
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP Mat_t(double* V, Long row, Long col) {
@@ -622,12 +626,11 @@ SEXP MatString(char **V, Long row, Long col, Long max) {
   if (V==NULL) return allocMatrix(STRSXP, 0, 0);
   Long n = row * col;
   if (n>max) return TooLarge(row, col);
-  SEXP dummy;
-  PROTECT(dummy=allocMatrix(STRSXP, row, col));
-  for (Long k=0; k<n; k++)
-    SET_STRING_ELT(dummy, k, mkChar(V[k]));
+  SEXP Ans;
+  PROTECT(Ans=allocMatrix(STRSXP, row, col));  
+  for (Long k=0; k<n; k++) SET_STRING_ELT(Ans, k, mkChar(V[k]));
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP MatString(char** V, Long row, Long col) {
@@ -638,11 +641,11 @@ SEXP MatInt(int* V, Long row, Long col, Long max) {
   if (V==NULL) return allocMatrix(INTSXP, 0, 0);
   Long n = row * col;
   if (n>max) return TooLarge(row, col);
-  SEXP dummy;
-  PROTECT(dummy=allocMatrix(INTSXP, row, col));
-  for (Long i=0; i<n; i++) INTEGER(dummy)[i] = V[i];
+  SEXP Ans;
+  PROTECT(Ans=allocMatrix(INTSXP, row, col));
+  MEMCOPY(INTEGER(Ans), V, n * sizeof(int));
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP MatInt(int* V, Long row, Long col) {
@@ -658,15 +661,14 @@ SEXP Array3D(double** V, Long depth, Long row, Long col, Long max) {
     int nn[3] = { (int) row, (int) col, (int) depth };
     return TooLarge(nn, 3);
   }
-  SEXP dummy;
-  PROTECT(dummy=alloc3DArray(REALSXP, depth, row, col));
-  for (Long j=0; j<depth; j++) {
-    for (Long i=0; i<m; i++) {
-      REAL(dummy)[j*m+i] = V[j][i];
-    }
-  }
+  SEXP Ans;
+  PROTECT(Ans=alloc3DArray(REALSXP, depth, row, col));
+  double *ans = REAL(Ans);
+  for (Long j=0; j<depth; j++) 
+    for (Long i=0; i<m; i++) 
+      ans[j*m+i] = V[j][i];
   UNPROTECT(1);
-  return dummy;
+  return Ans;
 }
 
 SEXP Array3D(double** V, Long depth, Long row, Long col) {
@@ -677,18 +679,21 @@ SEXP Array3D(double** V, Long depth, Long row, Long col) {
 
 
 usr_bool UsrBoolRelaxed(SEXP p, char *name, Long idx) {
-  double dummy = Real(p, name, idx);
-  if (!R_finite(dummy)) return Nan;
-  return dummy==0.0 ? False : True ;
+  double tmp = Real(p, name, idx);
+  if (!R_finite(tmp)) return Nan;
+  return tmp==0.0 ? False : True ;
 }
 
 
 usr_bool UsrBool(SEXP p, char *name, Long idx) {
-  double dummy = Real(p, name, idx);
-  if (dummy == 0.0) return False;
-  else if (dummy == 1.0) return True;
-  else if (ISNAN(dummy)) return Nan;
-  RFERROR2("invalid value (%d) for boolean variable '%.50s'.", (int) dummy, name);
+  double tmp = Real(p, name, idx);
+  usr_bool ans;
+  if (tmp == 0.0) ans= False;
+  else if (tmp == 1.0) ans= True;
+  else if (ISNAN(tmp)) ans= Nan;
+  else  RFERROR2("invalid value (%d) for boolean variable '%.50s'.",
+		 (int) tmp, name);
+  return ans;
 }
 
 
@@ -708,9 +713,7 @@ SEXP String(char V[][MAXCHAR], Long n, Long max) {
   if (n>max) return TooLarge(n);
   if (n<0) return TooSmall();
   PROTECT(str = allocVector(STRSXP, n)); 
-  for (Long i=0; i<n; i++) {
-    SET_STRING_ELT(str, i, mkChar(V[i]));
-  }
+  for (Long i=0; i<n; i++) SET_STRING_ELT(str, i, mkChar(V[i]));
   UNPROTECT(1);
   return str;
 }
@@ -723,12 +726,12 @@ SEXP String(int *V, const char * List[], Long n, Long endvalue) {
   if (V==NULL || n <= 0) return allocVector(STRSXP, 0);
   Long k;
   for (k=0; k<n; k++) {
+    //    printf("k=%d %d: %d %d\n", k,n, V[k], NoInversionMethod);
+    assert(V[k] <= endvalue);
     if (V[k] == endvalue) break;
   }
   PROTECT(str = allocVector(STRSXP, k)); 
-  for (Long i=0; i<k; i++) {
-    SET_STRING_ELT(str, i, mkChar(List[V[i]]));
-  }
+  for (Long i=0; i<k; i++) SET_STRING_ELT(str, i, mkChar(List[V[i]]));
   UNPROTECT(1);
   return str;
 }
@@ -749,7 +752,10 @@ double Real(SEXP p, char *name, Long idx) {
   }
 
   RFERROR2("'%.50s' can not be transformed to double! (got'%.50s')\n",
-	   name, R_TYPE_NAMES[TYPEOF(p)]);  
+	   name,
+	   TYPEOF(p) <= LAST_R_TYPE_NAME ? KKR_TYPE_NAMES[TYPEOF(p)] :
+	   "something else"
+	   );  
   return RF_NA;  // to avoid warning from compiler
 }
 
@@ -793,10 +799,13 @@ int Integer(SEXP p, char *name, Long idx, bool nulltoNA) {
     default : {}
     }
   } else if (nulltoNA) return NA_INTEGER;
-  
+
   RFERROR2("%.50s: incorrect type. Got '%.50s'.",
-	   name, R_TYPE_NAMES[TYPEOF(p)]);
-  return NA_INTEGER; // compiler warning vermeiden
+	   name,
+	   TYPEOF(p) <= LAST_R_TYPE_NAME ? KKR_TYPE_NAMES[TYPEOF(p)]
+	   : "something else");
+
+ return NA_INTEGER; // compiler warning vermeiden
 }
 
 int Integer(SEXP p, char *name, Long idx) {
@@ -819,15 +828,13 @@ void Integer(SEXP el, char *name, int *vec, Long maxn) {
 
 
 void Integer2(SEXP el, char *name, int *vec) {
-  Long n;
-  if (el == R_NilValue || (n = length(el))==0) {
-      RFERROR1("'%.50s' cannot be transformed to integer.\n",name);
-  }
+  Long n = length(el);
+  if (n == 0)  RFERROR1("'%.50s' cannot be transformed to integer.\n",name);
  
   vec[0] = Integer(el, name, 0);
   if (vec[0] != NA_INTEGER && vec[0] < 1)
     RFERROR1("first component of '%.50s' must be at least 1", name);
-  if (n==1) vec[1] = vec[0];
+  if (n == 1) vec[1] = vec[0];
   else {
     vec[1] = Integer(el, name, n-1);    
     if ( vec[1] != NA_INTEGER && vec[1] < vec[0])
@@ -869,7 +876,7 @@ char Char(SEXP el, char *name) {
   type = TYPEOF(el);
   if (type == CHARSXP) return CHAR(el)[0];
   if (type == STRSXP) {
-    if (length(el)==1) {
+    if (length(el) == 1) {
       if (STRLEN(CHAR(STRING_ELT(el,0))) == 1)
 	return (CHAR(STRING_ELT(el,0)))[0];
       else if (STRLEN(CHAR(STRING_ELT(el,0))) == 0)
@@ -938,10 +945,10 @@ double NonPosReal(SEXP el, char *name) {
 
 int PositiveInteger(SEXP el, char *name) {
   int num = INT;
-  if (num<=0) {
+  if (num <= 0) {
     WARN2("'%.50s', which has been %.50s, is set 1.\n",
-	  name, num==0L ? "0" : "negative");
-    num=1L;
+	  name, num ? "negative" : "0");
+    num=1;
   }
    return num;
 }
@@ -1045,18 +1052,17 @@ int Match(char *name, const char * List[], int n) {
 
 void GetName(SEXP el, char *name, const char * List[], int n,
 	     int defaultvalue, int endvalue, int *ans, int maxlen_ans) {
-  char dummy[1000];
-  int 
-    k = 0,
+  char msg0[1000];
+  int
+    k = 0, // globale variable !
     len_el = length(el);
 
-  if (TYPEOF(el) == NILSXP) goto ErrorHandling;
   if (len_el > maxlen_ans) 
     RFERROR2("option '%.50s' is too lengthy. Maximum length is %d.",
 	     name, maxlen_ans);
-
+  
   if (TYPEOF(el) == STRSXP) {    
-    for (k=0; k<len_el; k++) {
+    for (; k<len_el; k++) {
       ans[k] = Match((char*) CHAR(STRING_ELT(el, k)), List, n);
       if (ans[k] < 0) {
 	if (STRCMP((char*) CHAR(STRING_ELT(el, k)), " ") == 0 ||
@@ -1066,24 +1072,35 @@ void GetName(SEXP el, char *name, const char * List[], int n,
 	goto ErrorHandling0;
       }
     }
-    for (; k<maxlen_ans; k++) ans[k] = endvalue;
-    return;
+  } else {
+    Integer(el, name, ans, maxlen_ans);
+    for (; k<len_el; k++)
+      if (ans[k] < 0 || ans[k] >= n) goto ErrorHandling0;
   }
+ 
+  for (k=len_el; k<maxlen_ans; k++) ans[k] = endvalue;
+  return;
 
 ErrorHandling0:
-  SPRINTF(dummy, "'%.50s': unknown value '%.50s'. Possible values are:", 
-	  name, CHAR(STRING_ELT(el, k)));
+  if (TYPEOF(el) == STRSXP)
+    SPRINTF(msg0, "'%.50s': unknown value '%.50s'. Possible values are:", 
+	    name, CHAR(STRING_ELT(el, k)));
+  else 
+     SPRINTF(msg0,
+	     "'%.50s':  value '%d' not in {0,...%d}. Other possible values are:", 
+	     name, ans[k], n-1);
+  
   //  printf("'%s': ", CHAR(STRING_ELT(el, k)));
   int i;
   for (i=0; i<n-1; i++) {
     char msg[1000];
     // printf("'%s', ", List[i]);
-    SPRINTF(msg, "%.900s '%.50s',", dummy, List[i]);    
-    STRCPY(dummy, msg);
+    SPRINTF(msg, "%.900s '%.50s',", msg0, List[i]);    
+    STRCPY(msg0, msg);
   }
   //printf("'%s'\n ", List[i]);
   	  
-  RFERROR2("%.900s and '%.50s'.", dummy, List[i]);  
+  RFERROR2("%.900s and '%.50s'.", msg0, List[i]);  
  
  ErrorHandling:
   if (defaultvalue >= 0) {
